/*
 *  SSHAuth.cpp
 *
 *  Copyright (C) 2005-2006  Ahmed Obied
 *	
 *  This program is free software; you can redistribute it and/or	
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *  This product includes software developed by the libSSH2 Project
 *  (http://libssh2.sourceforge.net/)
 *
 *  This product includes software developed by the OpenSSL Project
 *  for use in the OpenSSL Toolkit. (http://www.openssl.org/)
 *
 *  In addition, as a special exception, the copyright holders give
 *  permission to link the code of portions of this program with the
 *  OpenSSL library under certain conditions as described in each
 *  individual source file, and distribute linked combinations
 *  including the two.
 *
 *  You must obey the GNU General Public License in all respects
 *  for all of the code used other than OpenSSL.  If you modify
 *  file(s) with this exception, you may extend this exception to your
 *  version of the file(s), but you are not obligated to do so.  If you
 *  do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source
 *  files in the program, then also delete it here.
 *
 */
#include "SSHAuth.h"
#include "resource.h"

HINSTANCE hInstance;

static PCHAR convert_wchar_to_char(PTCHAR wcstr)
{
	PCHAR cstr;
		
	cstr = (PCHAR)GlobalAlloc(NULL,MAX_BYTES);
	WideCharToMultiByte(CP_ACP,NULL,(LPCWSTR)wcstr,-1,cstr,MAX_BYTES,NULL,NULL);	
	return cstr;
}

static LPTSTR convert_char_to_wchar(PCHAR cstr)
{
	LPTSTR wcstr;
	INT wcstr_size;

	wcstr_size = lstrlenA(cstr);
	wcstr = (LPTSTR)SysAllocStringLen(NULL,wcstr_size);
	MultiByteToWideChar(CP_ACP,0,cstr,wcstr_size,wcstr,wcstr_size);
	return wcstr;
}

static LPTSTR read_registry_string(LPTSTR location)
{
	LONG l_result;
	HKEY h_key;
	DWORD bytes_read = MAX_BYTES;
	TCHAR result[MAX_BYTES];
	TCHAR *ret;
	
	l_result = RegOpenKeyEx(HKEY_LOCAL_MACHINE,REGISTRY_KEY,0,KEY_QUERY_VALUE,&h_key);
	if(l_result == ERROR_SUCCESS) {
    		l_result = RegQueryValueEx(h_key,location,NULL,NULL,(LPBYTE)&result,&bytes_read);
		if(l_result == ERROR_SUCCESS) {
			if(bytes_read > 0) {
				if(!(_tcscmp(result,EMPTY) == 0)) {
					ret = _tcsdup(result);
					RegCloseKey(h_key);
					return ret;
				}
			}
		}
	}
	RegCloseKey(h_key);
	return NULL;
}

static DWORD read_registry_dword(LPTSTR location)
{
	LONG l_result;
	HKEY h_key;
	DWORD value = 0;
	DWORD size = sizeof(DWORD);

	l_result = RegOpenKeyEx(HKEY_LOCAL_MACHINE,REGISTRY_KEY,0,KEY_QUERY_VALUE,&h_key);
	if(l_result == ERROR_SUCCESS) {
    		l_result = RegQueryValueEx(h_key,location,NULL,NULL,(LPBYTE)&value,&size);
			if(l_result == ERROR_SUCCESS) {
				RegCloseKey(h_key);
			        return value;
			}
	}
	RegCloseKey(h_key);
	return 0;
}

static BOOL write_registry_string(LPTSTR value, LPTSTR location)
{
	HKEY h_key;

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,REGISTRY_KEY,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,
		&h_key,NULL) == ERROR_SUCCESS) {
		if(value != NULL)
			if(RegSetValueEx(h_key,location,0,REG_SZ,(CONST BYTE *)value,
				(DWORD)(_tcslen(value)*sizeof(TCHAR))) == ERROR_SUCCESS) {
				RegFlushKey(h_key);
				RegCloseKey(h_key);
				return TRUE;
			 }
	}
	RegFlushKey(h_key);
	RegCloseKey(h_key);
	return TRUE;
}

static BOOL write_registry_dword(DWORD value, LPTSTR location)
{
	HKEY h_key;
	
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,REGISTRY_KEY,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,
		&h_key,NULL) == ERROR_SUCCESS) {
		if(RegSetValueEx(h_key,location,0,REG_DWORD,(LPBYTE)&value,sizeof(DWORD)) == ERROR_SUCCESS) {
			RegFlushKey(h_key);
			RegCloseKey(h_key);
			return TRUE;
		}
	}
	RegFlushKey(h_key);
	RegCloseKey(h_key);
	return FALSE;
}

static VOID load_data(HWND hwnd)
{
	HWND lhwnd;
	LPTSTR hosts, sentry;

	sentry = (LPTSTR)malloc(MAX_BYTES);
	lhwnd = GetDlgItem(hwnd,IDC_LIST_HOSTS);
	SendMessage(lhwnd,(UINT)LB_RESETCONTENT,(WPARAM)0,(LPARAM)0);
	if ((hosts = read_registry_string(REGISTRY_ENTRY)) != NULL) {
		DWORD i, j;
		
		i = 0;
		while (1) {
			if (i == _tcslen(hosts))
				break;
			memset(sentry,STR_TERM,MAX_BYTES);
			for (j = 0 ; i < _tcslen(hosts) ; i++, j++) {
				if (hosts[i] == _T(',')) {
					++i;				
					break;
				}
				sentry[j] = hosts[i];
			}
			SendMessage(lhwnd,(UINT)LB_ADDSTRING,(WPARAM)0,(LPARAM)sentry);
		}
	}
}

static LPTSTR set_host_entry(HWND hwnd, LPTSTR host, LPTSTR port)
{
	LPTSTR entry, hosts, sentry, centry;
	DWORD i, total;
	HWND lhwnd;

	centry = (LPTSTR)malloc(MAX_BYTES);
	sentry = (LPTSTR)malloc(MAX_BYTES);
	entry = (LPTSTR)malloc(MAX_PATH);
	memset(entry,STR_TERM,MAX_PATH);
	memset(sentry,STR_TERM,MAX_BYTES);
	hosts = read_registry_string(REGISTRY_ENTRY);
	if (hosts != NULL) {
		_tcscat(entry,hosts);	
		_tcscat(entry,_T(","));
	}
	_tcscat(sentry,host);
	_tcscat(sentry,_T(":"));
	_tcscat(sentry,port);
	_tcscat(entry,sentry);
	lhwnd = GetDlgItem(hwnd,IDC_LIST_HOSTS);
	total = SendMessage(lhwnd,(UINT)LB_GETCOUNT,(WPARAM)0,(LPARAM)0);
	for (i = 0 ; i < total ; i++) {
		memset(centry,STR_TERM,MAX_BYTES);
		SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)i,(LPARAM)centry);
		if (_tcscmp(sentry,centry) == 0) {
			error("You are trying to add a host entry that already exists.");
			return NULL;
		}
	}
	if (!write_registry_string(entry,REGISTRY_ENTRY)) {
		error("An error has occurred ... \n\nCannot save host data in the registry.");
		return NULL;
	}
	return entry;
}

static VOID add_host(HWND hwnd)
{
	LPTSTR host, port, entry; 
	DWORD dport;

	host = (LPTSTR)malloc(MAX_BYTES);
	port = (LPTSTR)malloc(MAX_BYTES);
	memset(host,STR_TERM,MAX_BYTES);
	memset(host,STR_TERM,MAX_BYTES);
	GetDlgItemText(hwnd,IDC_EDIT_HOST,host,MAX_BYTES);
	if (_tcscmp(host,EMPTY) == 0) {
		error("Please enter a valid host.");
		return;
	}
	GetDlgItemText(hwnd,IDC_EDIT_PORT,port,MAX_BYTES);
	if (_tcscmp(port,EMPTY) == 0) {
		error("Please enter a valid port.");
		return;
	}
	dport = atoi(convert_wchar_to_char(port));
	if (!((dport >= MIN_PORT_NUM) && (dport <= MAX_PORT_NUM))) {
		error("You have entered an invalid port.");
		return;
	}
	entry = set_host_entry(hwnd,host,port);
	if (entry == NULL)
		return;
	load_data(hwnd);
}

static VOID modify_host(HWND hwnd)
{
	HWND lhwnd;
	DWORD index, i, j, total;
	LPTSTR entry, sentry, centry;
	ENTRY_INFO entry_info;

	lhwnd = GetDlgItem(hwnd,IDC_LIST_HOSTS);
	index = SendMessage(lhwnd,(UINT)LB_GETCURSEL,(WPARAM)0,(LPARAM)0);
	if (index == LB_ERR) {
		error("Please select a host from the list to modify.");
		return;
	}
	centry = (LPTSTR)malloc(MAX_BYTES);
	sentry = (LPTSTR)malloc(MAX_BYTES);
	entry = (LPTSTR)malloc(MAX_PATH);
	entry_info.host = (LPTSTR)malloc(MAX_BYTES);
	entry_info.port = (LPTSTR)malloc(MAX_BYTES);
	memset(centry,STR_TERM,MAX_BYTES);
	memset(sentry,STR_TERM,MAX_BYTES);
	memset(entry,STR_TERM,MAX_PATH);
	memset(entry_info.host,STR_TERM,MAX_BYTES);
	memset(entry_info.port,STR_TERM,MAX_BYTES);
	total = SendMessage(lhwnd,(UINT)LB_GETCOUNT,(WPARAM)0,(LPARAM)0);
	SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)index,(LPARAM)centry);
	SendMessage(lhwnd,(UINT)LB_DELETESTRING,(WPARAM)index,(LPARAM)0);
	if (total != 1) {
		i = 0;
		while (1) {
			memset(sentry,STR_TERM,MAX_BYTES);
			SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)i,(LPARAM)sentry);
			_tcscat(entry,sentry);
			++i;
			if (i == (total - 1))
				break;
			_tcscat(entry,_T(","));
		}
		if (!write_registry_string(EMPTY,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot clear host data in the registry.");
			return;
		}
		if (!write_registry_string(entry,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot save host data in the registry.");
			return;
		}
	} else {
		if (!write_registry_string(EMPTY,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot clear host data in the registry.");
			return;
		}
	}
	for (i = 0, j = 0 ; ; i++, j++) {
		if (centry[i] == _T(':'))
			break;
		entry_info.host[j] = centry[i];
	}
	++i;
	for (j = 0 ; i < _tcslen(centry) ; i++, j++) 
		entry_info.port[j] = centry[i];
	SetDlgItemText(hwnd,IDC_EDIT_HOST,entry_info.host);
	SetDlgItemText(hwnd,IDC_EDIT_PORT,entry_info.port);
}

static VOID remove_host(HWND hwnd)
{
	HWND lhwnd;
	DWORD index, total, i;
	LPTSTR entry, sentry;

	lhwnd = GetDlgItem(hwnd,IDC_LIST_HOSTS);
	index = SendMessage(lhwnd,(UINT)LB_GETCURSEL,(WPARAM)0,(LPARAM)0);
	if (index == LB_ERR) {
		error("Please select a host from the list to remove.");
		return;
	}
	total = SendMessage(lhwnd,(UINT)LB_GETCOUNT,(WPARAM)0,(LPARAM)0);
	SendMessage(lhwnd,(UINT)LB_DELETESTRING,(WPARAM)index,(LPARAM)0);
	entry = (LPTSTR)malloc(MAX_PATH);
	sentry = (LPTSTR)malloc(MAX_BYTES);
	memset(entry,STR_TERM,MAX_PATH);
	if (total != 1) {
		i = 0;
		while (1) {
			memset(sentry,STR_TERM,MAX_BYTES);
			SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)i,(LPARAM)sentry);
			_tcscat(entry,sentry);
			++i;
			if (i == (total - 1))
				break;
			_tcscat(entry,_T(","));
		}
		if (!write_registry_string(EMPTY,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot clear host data in the registry.");
			return;
		}
		if (!write_registry_string(entry,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot save host data in the registry.");
			return;
		}
	} else {
		if (!write_registry_string(EMPTY,REGISTRY_ENTRY)) {
			error("An error has occurred ... \n\nCannot clear host data in the registry.");
			return;
		}
	}
}

static VOID save_authentication_type(HWND hwnd)
{
	DWORD auth_type;

	if (IsDlgButtonChecked(hwnd,IDC_RADIO_ALL) == BST_CHECKED)
		auth_type = __ALL;
	else
	if (IsDlgButtonChecked(hwnd,IDC_RADIO_ONE) == BST_CHECKED)
		auth_type = __ONE;
	if (!write_registry_dword(auth_type,REGISTRY_AUTH)) {
		error("An error has occurred ... \n\nCannot save authentication type in the registry.");
		return;
	}
}

static VOID set_authentication_type(HWND hwnd)
{
	DWORD auth_type;

	auth_type = read_registry_dword(REGISTRY_AUTH);
	switch (auth_type) {
		case __ONE:
			CheckRadioButton(hwnd,IDC_RADIO_ALL,IDC_RADIO_ONE,IDC_RADIO_ONE);
			break;
		default:
			CheckRadioButton(hwnd,IDC_RADIO_ALL,IDC_RADIO_ONE,IDC_RADIO_ALL);
	}
}

static VOID scroll_host(DWORD choice, HWND hwnd)
{
	HWND lhwnd;
	DWORD total, i, index;
	LPTSTR entry, sentry, centry;

	lhwnd = GetDlgItem(hwnd,IDC_LIST_HOSTS);
	index = SendMessage(lhwnd,(UINT)LB_GETCURSEL,(WPARAM)0,(LPARAM)0);
	if (index == LB_ERR) {
		error("Please select a host from the list first.");
		return;
	}
	total = SendMessage(lhwnd,(UINT)LB_GETCOUNT,(WPARAM)0,(LPARAM)0);
	if (total == 1) {
		error("There is only one host in the list.");
		return;
	}
	if ((index == 0 && choice == __UP) || (index == (total - 1) && choice == __DOWN)) {
		error("You can't change the order of the selected host.");	
		return;
	}
	entry = (LPTSTR)malloc(MAX_PATH);
	sentry = (LPTSTR)malloc(MAX_BYTES);
	centry = (LPTSTR)malloc(MAX_BYTES);
	memset(sentry,STR_TERM,MAX_BYTES);
	memset(entry,STR_TERM,MAX_PATH);
	SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)index,(LPARAM)sentry);
	SendMessage(lhwnd,(UINT)LB_DELETESTRING,(WPARAM)index,(LPARAM)0);
	switch (choice) {
		case __UP: 
			--index;
			break;
		case __DOWN:
			++index;
	}
	i = 0;
	while (1) {
		memset(centry,STR_TERM,MAX_BYTES);
		SendMessage(lhwnd,(UINT)LB_GETTEXT,(WPARAM)i,(LPARAM)centry);
		if (i == index && choice == __UP) {
			_tcscat(entry,sentry);
			_tcscat(entry,_T(","));
		}
		_tcscat(entry,centry);
		if (i == (index - 1) && choice == __DOWN) {
			_tcscat(entry,_T(","));
			_tcscat(entry,sentry);
		}
		++i;
		if (i == (total - 1))
			break;
		_tcscat(entry,_T(","));
	}
	if (!write_registry_string(EMPTY,REGISTRY_ENTRY)) {
		error("An error has occurred ... \n\nCannot clear host data in the registry.");
		return;
	}	
	if (!write_registry_string(entry,REGISTRY_ENTRY)) {
		error("An error has occurred ... \n\nCannot save host data in the registry.");
		return;
	}
	load_data(hwnd);
	SendMessage(lhwnd,(UINT)LB_SETCURSEL,(WPARAM)index,(LPARAM)0);
}

INT_PTR CALLBACK plugin_settings(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg) {
		case WM_INITDIALOG:
			set_authentication_type(hwnd);
			load_data(hwnd);
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wparam)) {
				case IDC_BUTTON_ADD:
					add_host(hwnd);
					return TRUE;
				case IDC_BUTTON_MODIFY:
					modify_host(hwnd);
					return TRUE;
				case IDC_BUTTON_REMOVE:
					remove_host(hwnd);
					return TRUE;
				case IDC_BUTTON_OK:
					save_authentication_type(hwnd);
					EndDialog(hwnd,LOWORD(wparam));
					return TRUE;
				case IDC_BUTTON_UP:
					scroll_host(__UP,hwnd);
					return TRUE;
				case IDC_BUTTON_DOWN:
					scroll_host(__DOWN,hwnd);
			}
			return TRUE;
		case WM_CLOSE:
			EndDialog(hwnd,LOWORD(wparam));
			return TRUE;
	}
	return FALSE;
}

BOOL APIENTRY DllMain(HANDLE handle, DWORD reason, LPVOID reserved)
{
	if (reason == DLL_PROCESS_ATTACH) {
		hInstance = (HINSTANCE)handle;
	}
	return TRUE;
}

SSHAUTH_API BOOL UserLogin(LPTSTR username, LPTSTR password, pGinaInfo *settingsInfo)
{
	BOOL auth_check;
	WSADATA wsa_data;
	DWORD err, port, auth_type, total, count;
	PCHAR user, pwd, str_port;
	LPTSTR hosts, sentry, host;
	ENTRY_INFO *entries, *oentries;

	auth_check = FALSE;
	user = convert_wchar_to_char(username);
	pwd = convert_wchar_to_char(password);
	str_port = (PCHAR)malloc(MAX_BYTES);
	sentry = (LPTSTR)malloc(MAX_BYTES);
	total = 0;
	auth_type = read_registry_dword(REGISTRY_AUTH);
	if ((hosts = read_registry_string(REGISTRY_ENTRY)) != NULL) {
		DWORD i, j, m, n;
		
		i = 0;
		while (1) {
			if (i == _tcslen(hosts))
				break;
			for (j = 0 ; i < _tcslen(hosts) ; i++) {
				if (hosts[i] == _T(',')) {
					++i;				
					break;
				}
			}
			++total;
		}
		entries = (ENTRY_INFO *)malloc(total * sizeof(ENTRY_INFO));
		oentries = entries;
		i = 0;
		while (1) {
			if (i == _tcslen(hosts))
				break;
			entries->host = (LPTSTR)malloc(MAX_BYTES);
			entries->port = (LPTSTR)malloc(MAX_BYTES);
			memset(entries->host,STR_TERM,MAX_BYTES);
			memset(entries->port,STR_TERM,MAX_BYTES);
			memset(sentry,STR_TERM,MAX_BYTES);
			for (j = 0 ; i < _tcslen(hosts) ; i++, j++) {
				if (hosts[i] == _T(',')) {
					++i;				
					break;
				}
				sentry[j] = hosts[i];
			}
			for (m = 0, n = 0 ; ; m++, n++) {
				if (sentry[m] == _T(':'))
					break;
				entries->host[n] = sentry[m];
			}
			++m;
			for (n = 0 ; m < _tcslen(sentry) ; m++, n++) 
				entries->port[n] = sentry[m];
			++entries;
		}
	}
	else {
		error("An error has occurred ... \n\nNo hosts are set.\n");
		return FALSE;
	}
	err = WSAStartup(WINSOCK_VERSION,&wsa_data);
	if (err != 0) {
		error("An error has occurred ... \n\nWSAStartup call failed.\n");
		return FALSE;
	}
	for (count = 0 ; count < total ; count++) {
		LIBSSH2_SESSION *ssh_session;
		ADDRINFO *addr_info;
		SOCKET sock_ds;

		if (auth_type == __ONE && auth_check == TRUE) 		
			break;
		auth_check = FALSE;
		host = oentries->host;
		port = atoi(convert_wchar_to_char(oentries->port));
		memset(str_port,STR_TERM,MAX_BYTES);
		sprintf(str_port,"%d",port);
		err = getaddrinfo(convert_wchar_to_char(host),str_port,NULL,&addr_info);
		if (err != 0) {
			if (auth_type == __ONE) {
				++oentries;				
				continue;
			}
			error("An error has occurred ... \n\ngetaddrinfo call failed.\n");
			break;
		}
		sock_ds = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
		if (sock_ds == INVALID_SOCKET) {
			if (auth_type == __ONE) {
				++oentries;				
				continue;
			}
			error("An error has occurred ... \n\nsocket call failed.\n");
			break;
		}
		err = connect(sock_ds,addr_info->ai_addr,addr_info->ai_addrlen);
		if (err == SOCKET_ERROR) {
			if (auth_type == __ONE) {
				++oentries;				
				continue;
			}
			error("An error has occurred ... \n\nconnect call failed.\n");
			break;
		}
		ssh_session = libssh2_session_init();
		if (ssh_session == NULL) {
			if (auth_type == __ONE) {
				++oentries;				
				continue;
			}
			error("An error has occurred ... \n\nlibssh2_session_init call failed.\n");
			break;
		}
		err = libssh2_session_startup(ssh_session,(int)sock_ds);
		if (err != 0) {
			if (auth_type == __ONE) {
				++oentries;				
				continue;
			}
			error("An error has occurred ...\n\nlibssh2_session_startup call failed.\n");
			break;
		}
		err = libssh2_userauth_password(ssh_session,user,pwd);
		if (err != 0)
			auth_check = FALSE;
		else
			auth_check = TRUE;
		libssh2_session_free(ssh_session);
		closesocket(sock_ds);
		++oentries;
	}
	return auth_check;
}

SSHAUTH_API BOOL ChangeUserPassword(LPTSTR username, LPTSTR old_pwd, LPTSTR new_pwd)
{
	return FALSE;
}

SSHAUTH_API LPCTSTR AboutPlugin()
{
	LPCTSTR about = _T("SSH Authentication plugin for pGINA by Ahmed Obied (ahmed.obied@gmail.com)");
	return about;
} 

SSHAUTH_API VOID ChangePluginSettings()
{
	DialogBox(hInstance,MAKEINTRESOURCE(IDD_CONFIG_DIALOG),NULL,plugin_settings);	
}

SSHAUTH_API VOID LoginHook(pGinaInfo *settings_info)
{
	
}

SSHAUTH_API BOOL IsRequired()
{
	return FALSE;
}

SSHAUTH_API VOID LogoutHook(pGinaInfo *settings_info)
{

}
