/*
 *  SSHAuth.h
 *
 *  Copyright (C) 2005-2006  Ahmed Obied
 *	
 *  This program is free software; you can redistribute it and/or	
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *  This product includes software developed by the libSSH2 Project
 *  (http://libssh2.sourceforge.net/)
 *
 *  This product includes software developed by the OpenSSL Project
 *  for use in the OpenSSL Toolkit. (http://www.openssl.org/)
 *
 *  In addition, as a special exception, the copyright holders give
 *  permission to link the code of portions of this program with the
 *  OpenSSL library under certain conditions as described in each
 *  individual source file, and distribute linked combinations
 *  including the two.
 *
 *  You must obey the GNU General Public License in all respects
 *  for all of the code used other than OpenSSL.  If you modify
 *  file(s) with this exception, you may extend this exception to your
 *  version of the file(s), but you are not obligated to do so.  If you
 *  do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source
 *  files in the program, then also delete it here.
 *
 */

#ifndef _SSH_AUTH_HEADER_
#define _SSH_AUTH_HEADER_

#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "libssh2/libssh2.h"

#ifdef SSHAUTH_EXPORTS
#define SSHAUTH_API extern "C" __declspec(dllexport)
#else
#define SSHAUTH_API extern "C" __declspec(dllimport)
#endif

#define MAX_BYTES		1024
#define MIN_PORT_NUM		1
#define MAX_PORT_NUM		65535
#define REG_DATA_NUM		3
#define REGISTRY_ENTRY		_T("Hosts")
#define REGISTRY_AUTH		_T("Authentication")
#define REGISTRY_KEY		_T("SOFTWARE\\pGina\\SSHAuth")
#define STR_TERM		_T('\0')
#define EMPTY			_T("")
#define NAME			_T("SSH Authentication plugin")

#define error(x) { \
	MessageBox(GetForegroundWindow(),_T(x),NAME,MB_ICONERROR); \
}

enum AUTHENTICATION_ID
{
	__ALL = 0,
	__ONE = 1
};

enum SORT_ID
{
	__UP = 0,
	__DOWN = 1
};

typedef struct {
	LPTSTR host;
	LPTSTR port;
} ENTRY_INFO;

typedef struct pGinaInfo {
	LPTSTR pathMSGina;
	LPTSTR pathPlugin;
	LPTSTR pathProfile;
	LPTSTR mapPaths;
	LPTSTR Username;
	LPTSTR Password;
	LPTSTR homeDir;
	BOOL isAdmin;
	BOOL disabled;
	INT authType;
	HANDLE hUser;
	LPTSTR userGroups;
	LPTSTR userDescription;
	LPTSTR userFullName;
	BOOL allowPassChange;
	LPTSTR errorString;
	LPTSTR defaultDomain;
	BOOL Reserved3;
	BOOL Reserved4;
} pGinaInfo;

SSHAUTH_API BOOL UserLogin(LPTSTR,LPTSTR,pGinaInfo *);			
SSHAUTH_API BOOL ChangeUserPassword(LPTSTR,LPTSTR,LPTSTR);
SSHAUTH_API LPCTSTR AboutPlugin();
SSHAUTH_API VOID ChangePluginSettings();
SSHAUTH_API VOID LoginHook(pGinaInfo *);
SSHAUTH_API VOID LogoutHook(pGinaInfo *);
SSHAUTH_API BOOL IsRequired();

#endif